/*
 * block-compressor.hpp for bwt tunneling
 * Copyright (c) 2018 Uwe Baier, Kadir Dede All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#ifndef _BLOCK_COMPRESSOR_HPP
#define _BLOCK_COMPRESSOR_HPP

#include <assert.h>
#include <forward_list>
#include <ios>
#include <iostream>
#include <istream>
#include <iterator>
#include <limits>
#include <ostream>
#include <sstream>
#include <stdexcept>
#include <string>

//! abstract base class for a block compressor.
/*! a block compressor divides its input into blocks
   (a continous sequence of characters of the input) and compresses
   each block individually.
 */
class block_compressor {
	private:
		//current block size (it is ensured that block size is smaller than maxblocksize)
		std::streamsize blocksize;
		// constant indicating how big a block can maximally be
		const std::streamsize maxblocksize;
		bool quiet = true; //indicates whether compressor is quiet and does not print any additional information
	protected:
		virtual void check_block_threshold( std::istream &in, std::streampos end, std::string filename ) const = 0;

		//a function to print information during encoding (function will not print
		// something if compressor is set to be quiet, what is the default)
		template<class V>
		void print_info( std::string key, V value ) const {
			if (!quiet) {
				std::cout << key << value;
			} //TODO: nachher weitermachen
		};

	public:
		//! constructor, expects maximal block size possible.
		block_compressor( std::streamsize max_block_size )
		                : blocksize( max_block_size ), maxblocksize( max_block_size ) {};

		//! sets the quiet state of this compressor (quiet=true is default).
		/*! if the compressor is set to quiet, it will not print any extra information,
		   otherwise it will print extra information related to the compression method used.
		 */
		void set_quiet( bool q ) {
			quiet = q;
		};

		//! returns whether this compressor is quiet (see set_quiet).
		bool is_quiet() const {
			return quiet;
		};

		//! returns current block size. Block size initially is set to the maximal
		//! possible block size.
		std::streamsize get_block_size() const {
			return blocksize;
		};

		//! returns maximal possible block size
		std::streamsize get_max_block_size() const {
			return maxblocksize;
		};

		//! can be used to set the block size, value must be smaller
		//! or equal to get_max_block_size().
		void set_block_size( std::streamsize bs ) {
			assert( bs > 0 && bs <= get_max_block_size() );
			blocksize = bs;
		};

		//! checks the threshold
		void check_threshold( std::istream &in, std::string filename ) const {
			//set exception mask of instream
			in.exceptions( std::istream::badbit | std::istream::eofbit );

			//get length of input
			in.seekg(0, std::ios_base::end);
			std::streamsize n = in.tellg();
			in.seekg(0, std::ios_base::beg); //jump to start of stream again

			//process each block
			while (n > 0) {
				auto bs = std::min(n, get_block_size());
				check_block_threshold( in, in.tellg()+bs, filename );
				n -= bs;
			}				
		};

		//! checks the threshold
		void check_threshold( const std::string &S ) const {
			std::istringstream in( S );
			check_threshold( in, "string" );
		};

		//! utility for writing POD types to a stream.
		template<class T>
		static void write_primitive( T p, std::ostream &out ) {
			for (size_t i = 0; i < sizeof(T); i++) {
				out.put( (char)(p & std::numeric_limits<unsigned char>::max()) );
				p >>= std::numeric_limits<unsigned char>::digits;
			}
		};

		//! utility for reading POD types from a stream.
		template<class T>
		static T read_primitive( std::istream &in ) {
			T p = (T)0;
			for (size_t i = 0; i < sizeof(T); i++) {
				p |= (T)( (unsigned char)in.get() ) << 
					( i * std::numeric_limits<unsigned char>::digits );
			}
			return p;
		};
};

#endif
