/*
 * tbwt-compressor.hpp for bwt tunneling
 * Copyright (c) 2018 Uwe Baier, Kadir Dede All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#ifndef _TBWT_COMPRESSOR_HPP
#define _TBWT_COMPRESSOR_HPP

#include "block-compressor.hpp"

#include "aux-encoding.hpp"
#include "bwt-config.hpp"
#include "bwt-run-support.hpp"
#include "divsufsort.h"
#include "lheap.hpp"
#include "tunneling-support.hpp"

#include <array>
#include <assert.h>
#include <chrono>
#include <iterator>
#include <limits>
#include <stdexcept>
#include <stdint.h>
#include <string>
#include <utility>
#include <vector>
#include <sstream>

#define THRESHOLD_MAX 50

//! a tunneled-bwt-based generic compressor
template<class t_2st_encoder>
class tbwt_compressor : public block_compressor {
	protected:
		virtual void check_block_threshold( std::istream &in, std::streampos end, std::string filename ) const;
	public:
		//! constructor
		tbwt_compressor() : block_compressor( t_max_size ) {};
};

//// COMPRESSION //////////////////////////////////////////////////////////////

template<class t_ss_e>
void tbwt_compressor<t_ss_e>::check_block_threshold( std::istream &in, std::streampos end, std::string filename ) const {
	using namespace std;
	using namespace std::chrono;
	typedef typename istream::char_type schar_t;
	typedef high_resolution_clock timer;
	static_assert( is_same<
	                    typename make_unsigned<schar_t>::type,
	                    typename make_unsigned<t_uchar_t>::type
	               >::value,
	               "character types must be compatible" );

	//// GET INPUT ////////////////////////////////////////////////////////

	//get length of input
	t_size_t n = (t_size_t)(end - in.tellg());
	assert(n <= t_max_size );
	//print_info("input size", n);

	//read string from input and reverse it
	t_string_t S( n );
	in.read( (schar_t *)S.data(), n );

	//// BW-TRANSFORM INPUT ///////////////////////////////////////////////

	auto start = timer::now();
	saidx_t bwt_idx = 0;
	if (bw_transform(S.data(), S.data(), NULL, (saidx_t)n, &bwt_idx) < 0) {
		throw runtime_error( string("BW Transformation failed") );
	}
	auto stop = timer::now();
	//print_info("bwt construction time", (uint64_t)duration_cast<milliseconds>( stop - start ).count() );
	
	//// SET UP BWT NAVIGATION ////////////////////////////////////////////

	start = timer::now();
	bwt_run_support bwtrs( S.data(), n, bwt_idx );
	
	//// COMPUTE BLOCKS AND COLLISIONS ////////////////////////////////////
	
	tunneling_support<t_ss_e> ts( bwtrs );
	stop = timer::now();
	//print_info("block computation time", (uint64_t)duration_cast<milliseconds>( stop - start ).count() );
	
	//// SET UP AN ARRAY CONTAINING ALL USABLE BLOCKS /////////////////////
	
	vector<t_idx_t> H;
	H.reserve( ts.bns.blocks ); //put all blocks to a heap which are worth to be tunneled
	for (t_idx_t b = 0; b < ts.bns.blocks; b++) {
		if (ts.blockstate(b) != tunneling_support<t_ss_e>::CLEARED) {
			H.push_back( b );
		}
	}
	//output stats header
	print_info("file;n;r;rg1;rc", "");
	for (int k = 0; k <= THRESHOLD_MAX; k++) {
		print_info(";", k);
	}
	//output stats
	print_info( "\n", filename );
	print_info(";", n );
	print_info(";", bwtrs.runs );
	print_info(";", ts.tes.get_runs_hg1());
	print_info(";", ts.tes.get_rc());

	//// COMPUTE STATS ////////////////////////////////////////////////////
	for(int k = 0; k <= THRESHOLD_MAX; k++) {

		vector<t_idx_t> TB; //blocks to be tunneled
		TB.reserve( H.size() );

		for (t_idx_t b = 0; b < H.size(); b++) {
			if (ts.tes.blockscore( H[b] ) >= k) {
				TB.push_back( H[b] );
			}
		}

		//// TUNNEL BEST CHOICE OF BLOCKS /////////////////////////////////////
		t_string_t Scpy = S;
		tunneling_support<t_ss_e> tscpy = ts;
		twobitvector aux; //auxiliary structure for tunneling

		auto tbwt_idx = tscpy.tunnel_bwt( Scpy, aux, TB.begin(), TB.end() );
		stop = timer::now();

		//// WRITE ENCODING ///////////////////////////////////////////////////

		start = timer::now();
		t_ss_e::transform_aux( Scpy, tbwt_idx, aux );

		ostringstream sout;
		auto tbwencstartpos = sout.tellp();
		t_ss_e::encode( Scpy, sout );
		t_ss_e::encode( aux, sout );
		auto tbwtencendpos = sout.tellp();

		stop = timer::now();
		print_info(";", tbwtencendpos - tbwencstartpos );
	}
}

#endif
