/*
 * ui.cpp for bwt tunneling
 * Copyright (c) 2018 Uwe Baier, Kadir Dede All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#include <fstream>
#include <iostream>
#include <stdexcept>
#include <string>
#include <string.h>

#if defined TBWT
	#include "bw94-compressor.hpp"
	#define FILESUFFIX ".tbwz"
	#define COMPRESSOR tbwt_compressor_bw94
#elif defined TBCM
	#include "bcm-compressor.hpp"
	#define FILESUFFIX ".tbcm"
	#define COMPRESSOR tbwt_compressor_bcm
#elif defined TWT
	#include "wt-compressor.hpp"
	#define FILESUFFIX ".twt"
	#define COMPRESSOR tbwt_compressor_wt
#else
	#error unknown block compressor
#endif

using namespace std;

const int MODE_COMPRESS = 0;
const int MODE_DECOMPRESS = 1;

void printUsage(const char *cmd) {
	cerr << "usage: " << cmd << " INFILE" << endl;
	cerr << "\tINFILE: file to be checked for threshold" << endl;
	cerr << "\tProgram outputs compressed size in dependency of used threshold" << endl;
	cerr << "\tincluding a header line with more information" << endl;
}

int main( int argc, char **argv ) {
	//analyse args
	string infile;
	if (argc != 2) {
		printUsage(argv[0]);
		return 1;
	}
	infile = argv[1];
	// thanks to https://stackoverflow.com/questions/8520560/get-a-file-name-from-a-path
	string base_infile = infile.substr(infile.find_last_of("/\\") + 1);

	//open stream for infile
	ifstream fin{ infile };
	if (!fin) {
		printUsage(argv[0]);
		cerr << "unable to open file \"" << infile << "\"" << endl;
		return 1;
	}

	//compress or decompress, depending on mode
	COMPRESSOR compressor;
	compressor.set_quiet(false);
	try {
		compressor.check_threshold( fin, base_infile );
	} catch ( invalid_argument &e ) {
		printUsage( argv[0] );
		cerr << "Invalid argument: " << e.what() << endl;
		return 1;
	} catch ( runtime_error &e ) {
		printUsage( argv[0] );
		cerr << "Runtime error: " << e.what() << endl;
		return 1;
	} catch ( exception &e ) {
		cerr << "Exception: " << e.what() << endl;
		return 1;
	}
	return 0;	
}
